from logging import Handler
from queue import Queue
import sys
from threading import Thread
import logging
import asyncio
from functools import partial, wraps

RED = '\033[91m'
BLUE = '\033[94m'
BOLD = '\033[1m'
END = '\033[0m'


def patch_async_emit(handler: Handler):
    base_emit = handler.emit
    queue = Queue()

    def loop():
        while True:
            record = queue.get()
            try:
                base_emit(record)
            except:
                print(sys.exc_info())

    def async_emit(record):
        queue.put(record)

    thread = Thread(target=loop)
    thread.daemon = True
    thread.start()
    handler.emit = async_emit
    return handler


def patch_add_handler(logger):
    base_add_handler = logger.addHandler

    def async_add_handler(handler):
        async_handler = patch_async_emit(handler)
        base_add_handler(async_handler)

    return async_add_handler


def setup():
    logger = logging.getLogger()
    logger.handlers = []
    logger.addHandler = patch_add_handler(logger)
    handler = patch_async_emit(logging.StreamHandler())
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    handler.setFormatter(formatter)
    logger.addHandler(handler)


def log(fn=None, logger=logging.getLogger(), debug_level=logging.DEBUG):
    """
    logs parameters and result - takes no arguments
    """
    if fn is None:
        return partial(log, logger=logger, debug_level=debug_level)

    @wraps(fn)
    def func(*args, **kwargs):
        arg_string = ""
        for i in range(0, len(args)):
            var_name = fn.__code__.co_varnames[i]
            if var_name not in ['self', 'cls']:
                arg_string += var_name + ":" + str(args[i]) + ","
        arg_string = arg_string[0:len(arg_string) - 1]
        string = (RED + BOLD + '>> ' + END + 'Calling {0}({1})'.format(fn.__name__, arg_string))
        if len(kwargs):
            string = (
                RED + BOLD + '>> ' + END + 'Calling {0} with args {1} and kwargs {2}'.format(fn.__name__, arg_string,
                                                                                             kwargs))
        logger.log(debug_level, string)
        wrapped_fn = fn
        if not asyncio.iscoroutinefunction(fn):
            wrapped_fn = asyncio.coroutine(fn)
        try:
            result = yield from wrapped_fn(*args, **kwargs)
            string = BLUE + BOLD + '<< ' + END + 'Return {0} with result :{1}'.format(fn.__name__, result)
            logger.log(debug_level, string)
            return result
        except Exception as e:
            string = (RED + BOLD + '>> ' + END + '{0} raised exception :{1}'.format(fn.__name__, str(e)))
            logger.log(debug_level, string)
            raise e

    return func


def logx(supress_args=[], supress_all_args=False, supress_result=False, logger=logging.getLogger(),
         debug_level=logging.DEBUG):
    """
    logs parameters and result
    takes arguments
        supress_args - list of parameter names to supress
        supress_all_args - boolean to supress all arguments
        supress_result - boolean to supress result
        receiver - custom logging function which takes a string as input; defaults to logging on stdout
    """

    def decorator(fn):
        def func(*args, **kwargs):
            if not supress_all_args:
                arg_string = ""
                for i in range(0, len(args)):
                    var_name = fn.__code__.co_varnames[i]
                    if var_name != "self" and var_name not in supress_args:
                        arg_string += var_name + ":" + str(args[i]) + ","
                arg_string = arg_string[0:len(arg_string) - 1]
                string = (RED + BOLD + '>> ' + END + 'Calling {0}({1})'.format(fn.__name__, arg_string))
                if len(kwargs):
                    string = (
                        RED + BOLD + '>> ' + END + 'Calling {0} with args {1} and kwargs {2}'.format(
                            fn.__name__,
                            arg_string, kwargs))
                logger.log(debug_level, string)

            wrapped_fn = fn
            if not asyncio.iscoroutinefunction(fn):
                wrapped_fn = asyncio.coroutine(fn)
            result = yield from wrapped_fn(*args, **kwargs)

            if not supress_result:
                string = BLUE + BOLD + '<< ' + END + 'Return {0} with result : {1}'.format(fn.__name__, result)
                logger.log(debug_level, string)
            return result

        return func

    return decorator





from asyncio import Future
import logging
import random
from collections import defaultdict

from again.utils import unique_hex

from .services import TCPServiceClient

class RegistryClient:

    logger = logging.getLogger(__name__)

    def __init__(self, loop, host, port, bus):
        self._bus = bus
        self._loop = loop
        self._host = host
        self._port = port
        self._transport = None
        self._protocol = None
        self._service = None
        self._version = None
        self._node_id = None
        self._pending_requests = {}
        self._available_services = defaultdict(list)
        self._assigned_services = defaultdict(lambda: defaultdict(list))

    def _register_service(self, ip, port, service, vendors, version, type):
        self._service = service
        self._version = version
        packet = self._make_registration_packet(ip, port, service, version, vendors, type)
        self._protocol.send(packet)

    def add_service_death_listener(self, service, version):
        packet = self._make_death_subscription_packet(service, version, self._node_id)
        self._protocol.send(packet)

    def get_instances(self, service, version):
        packet = self._make_get_instance_packet(service, version)
        self._protocol.send(packet)

    def register_http(self, vendors, ip, port, service, version):
        self._register_service(ip, port, service, vendors, version, 'http')

    def register_tcp(self, vendors, ip, port, service, version):
        self._register_service(ip, port, service, vendors, version, 'tcp')
        self._register_for_subscription(vendors, ip, port)

    def subscribe_for_message(self, packet):
        packet['node_id'] = self._node_id
        self._protocol.send(packet)

    def _protocol_factory(self):
        from vyked.jsonprotocol import RegistryClientProtocol

        p = RegistryClientProtocol(self)
        return p

    def connect(self):
        coro = self._loop.create_connection(self._protocol_factory, self._host, self._port)
        self._transport, self._protocol = self._loop.run_until_complete(coro)

    def receive(self, packet:dict, registry_protocol):
        if packet['type'] == 'registered':
            self.cache_vendors(packet['params']['vendors'])
            self._bus.registration_complete()
        elif packet['type'] == 'deregister':
            self._handle_deregistration(packet)
        elif packet['type'] == 'subscription_list':
            self._handle_subscription_list(packet)
        elif packet['type'] == 'message_subscription_list':
            self._handle_subscription_list(packet)

    def get_all_addresses(self, full_service_name):
        return self._available_services.get(
            self._get_full_service_name(full_service_name[0], full_service_name[1]))

    def get_random_service(self, service_name):
        services = self._available_services[service_name]
        if len(services):
            return random.choice(services)
        else:
            return None

    def resolve(self, service: str, version: str, entity:str):
        service_name = self._get_full_service_name(service, version)
        if entity is not None:
            entity_map = self._assigned_services.get(service_name)
            if entity_map is None:
                self._assigned_services[service_name] = {}
            entity_map = self._assigned_services.get(service_name)
            if entity in entity_map:
                return entity_map[entity]
            else:
                host, port, node_id = self.get_random_service(service_name)
                if node_id is not None:
                    entity_map[entity] = host, port, node_id
                return host, port, node_id
        else:
            return self.get_random_service(service_name)

    def _make_registration_packet(self, ip:str, port:str, service:str, version:str, vendors, type:str):
        vendors_list = []
        for vendor in vendors:
            vendor_dict = {'service': vendor.name,
                           'version': vendor.version}
            vendors_list.append(vendor_dict)
        self._node_id = unique_hex()
        params = {'service': service,
                  'version': version,
                  'host': ip,
                  'port': port,
                  'node_id': self._node_id,
                  'vendors': vendors_list,
                  'type': type}
        packet = {'pid': unique_hex(),
                  'type': 'register',
                  'params': params}
        return packet

    def _make_death_subscription_packet(self, service:str, version:str, node_id:str):
        params = {'service': service,
                  'version': version}
        packet = {'pid': unique_hex(),
                  'type': 'service_death_sub',
                  'service': self._service,
                  'version': self._version,
                  'params': params,
                  'node': node_id}
        return packet

    def _make_get_instance_packet(self, service, version):
        params = {'service': service,
                  'version': version}
        packet = {'pid': unique_hex(),
                  'type': 'get_instances',
                  'service': self._service,
                  'version': self._version,
                  'params': params}
        return packet

    @staticmethod
    def _get_full_service_name(service, version):
        return "{}/{}".format(service, version)

    def cache_vendors(self, vendors):
        for vendor in vendors:
            vendor_name = vendor['name']
            for address in vendor['addresses']:
                self._available_services[vendor_name].append((address['host'], address['port'], address['node_id']))

    def _register_for_subscription(self, vendors, ip, port):
        subscription_packet = {
            'type': 'subscribe',
        }
        params = {
            'ip': ip,
            'port': port,
            'node_id': self._node_id
        }
        subscription_list = []
        for vendor in vendors:
            if isinstance(vendor, TCPServiceClient):
                for each in dir(vendor):
                    fn = getattr(vendor, each)
                    if callable(fn) and getattr(fn, 'is_subscribe', False):
                        subscription_list.append({
                            'service': vendor.name,
                            'version': vendor.version,
                            'endpoint': fn.__name__
                        })
        params['subscribe_to'] = subscription_list
        subscription_packet['params'] = params
        self._protocol.send(subscription_packet)

    def resolve_publication(self, service, version, endpoint):
        future = Future()
        request_id = unique_hex()
        packet = {'type': 'resolve_publication', 'request_id': request_id}
        params = {
            'service': service,
            'version': version,
            'endpoint': endpoint
        }
        packet['params'] = params
        self._pending_requests[request_id] = future
        self._protocol.send(packet)
        return future

    def resolve_message_publication(self, service, version, endpoint, entity):
        future = Future()
        request_id = unique_hex()
        packet = {'type': 'resolve_message_publication', 'request_id': request_id}
        params = {
            'service': service,
            'version': version,
            'endpoint': endpoint,
            'entity': entity
        }
        packet['params'] = params
        self._pending_requests[request_id] = future
        self._protocol.send(packet)
        return future

    def _handle_subscription_list(self, packet):
        future = self._pending_requests.pop(packet['request_id'])
        future.set_result(packet['nodes'])

    def _handle_deregistration(self, packet):
        params = packet['params']
        vendor = params['vendor']
        node = params['node_id']
        for each in self._available_services[vendor]:
            if each[2] == node:
                self._available_services[vendor].remove(each)
        entity_map = self._assigned_services.get(vendor)
        if entity_map is not None:
            stale_entities = []
            for entity, node_id in entity_map.items():
                if node == node_id:
                    stale_entities.append(entity)
            for entity in stale_entities:
                entity_map.pop(entity)









import asyncio
import json
import logging

from jsonstreamer import ObjectStreamer

from .registry import Registry
from .registryclient import RegistryClient
from .services import TCPServiceClient


class JSONProtocol(asyncio.Protocol):

    logger = logging.getLogger(__name__)

    def __init__(self):
        self._pending_data = []
        self._connected = False
        self._transport = None
        self._obj_streamer = None

    def _make_frame(self, packet):
        string = json.dumps(packet) + ','
        return string

    def _write_pending_data(self):
        for packet in self._pending_data:
            frame = self._make_frame(packet)
            self._transport.write(frame.encode())

    def connection_made(self, transport):
        self._connected = True
        self._transport = transport
        self._obj_streamer = ObjectStreamer()
        self._obj_streamer.auto_listen(self, prefix='on_')

        self._transport.write('['.encode())  # start a json array
        self._write_pending_data()

    def connection_lost(self, exc):
        self._connected = False
        self.logger.debug('Peer closed the connection')

    def send(self, packet: dict):
        string = json.dumps(packet) + ','
        if self._connected:
            self._transport.write(string.encode())
            self.logger.debug('Data sent: {}'.format(string))
        else:
            self._pending_data.append(packet)
            self.logger.debug('Appended data: {}'.format(self._pending_data))

    def close(self):
        self._transport.write('bye]'.encode())  # end the json array
        self._transport.close()

    def data_received(self, byte_data):
        string_data = byte_data.decode()
        self.logger.debug('Data received: {}'.format(string_data))
        self._obj_streamer.consume(string_data)

    def on_object_stream_start(self):
        raise RuntimeError('Incorrect JSON Streaming Format: expect a JSON Array to start at root, got object')

    def on_object_stream_end(self):
        del self._obj_streamer
        raise RuntimeError('Incorrect JSON Streaming Format: expect a JSON Array to end at root, got object')

    def on_array_stream_start(self):
        self.logger.debug('Array Stream started')

    def on_array_stream_end(self):
        del self._obj_streamer
        self.logger.debug('Array Stream ended')

    def on_pair(self, pair):
        self.logger.debug('Pair {}'.format(pair))
        raise RuntimeError('Received a key-value pair object - expected elements only')


class ServiceHostProtocol(JSONProtocol):
    def __init__(self, bus):
        super(ServiceHostProtocol, self).__init__()
        self._bus = bus

    def connection_made(self, transport):
        peername = transport.get_extra_info('peername')
        self.logger.debug('Client Connection from {}'.format(peername))
        super(ServiceHostProtocol, self).connection_made(transport)

    def connection_lost(self, exc):
        super(ServiceHostProtocol, self).connection_lost(exc)
        # TODO: think about what needs to be done here

    def on_element(self, element):
        self._bus.host_receive(packet=element, protocol=self)


class ServiceClientProtocol(JSONProtocol):
    def __init__(self, bus):
        super(ServiceClientProtocol, self).__init__()
        self._bus = bus

    def set_service_client(self, service_client:TCPServiceClient):
        self._service_client = service_client

    def connection_made(self, transport):
        peername = transport.get_extra_info('peername')
        self.logger.debug('Connected to server{}'.format(peername))
        super(ServiceClientProtocol, self).connection_made(transport)

    def on_element(self, element):
        self._bus.client_receive(packet=element, service_client=self._service_client)


class RegistryProtocol(JSONProtocol):
    def __init__(self, registry:Registry):
        super(RegistryProtocol, self).__init__()
        self._registry = registry

    def connection_made(self, transport):
        peername = transport.get_extra_info('peername')
        self._transport = transport
        self.logger.debug('Connected from {}'.format(peername))
        super(RegistryProtocol, self).connection_made(transport)
        # TODO: pass protocol to registry

    def on_element(self, element):
        self._registry.receive(packet=element, registry_protocol=self, transport=self._transport)


class RegistryClientProtocol(JSONProtocol):
    def __init__(self, registry_client:RegistryClient):
        super(RegistryClientProtocol, self).__init__()
        self._registry_client = registry_client

    def connection_made(self, transport):
        peername = transport.get_extra_info('peername')
        self.logger.debug('Connected to server{}'.format(peername))
        super(RegistryClientProtocol, self).connection_made(transport)

    def on_element(self, element):
        self._registry_client.receive(packet=element, registry_protocol=self)



class Pinger:

    def __init__(self, registry, loop, protocol, node):
        self._registry = registry
        self._protocol = protocol
        self._node = node
        self._loop = loop
        self._count = 0
        self._timer = None

    def start_ping(self):
        packet = self._make_ping_packet()
        self._loop.call_later(10, self._send_timed_ping, packet)

    def pong_received(self, count):
        if self._count == count:
            self._count += 1
            if self._timer is not None:
                self._timer.cancel()
        else:
            self._ping_timed_out()

    def _make_ping_packet(self):
        packet = {'type': 'ping', 'node_id': self._node, 'count': self._count}
        return packet

    def _send_timed_ping(self, packet):
        self._protocol.send(packet)
        self._timer = self._loop.call_later(5, self._ping_timed_out)

    def _ping_timed_out(self):
        self._timer.cancel()
        self._registry.handle_ping_timeout(self._node)


__all__ = ['TCPServiceClient', 'TCPApplicationService', 'TCPDomainService', 'TCPInfraService', 'HTTPServiceClient',
           'HTTPApplicationService', 'HTTPDomainService', 'HTTPInfraService', 'api', 'request', 'subscribe', 'publish',
           'message_pub', 'get', 'post', 'head', 'put', 'patch', 'delete', 'options', 'trace', 'Entity', 'Value',
           'Aggregate', 'Factory', 'Repository', 'Bus', 'Registry', 'PostgresStore', 'log', 'cursor', 'dict_cursor',
           'nt_cursor', 'transaction']

from .services import (TCPServiceClient, TCPApplicationService, TCPDomainService, TCPInfraService, HTTPServiceClient,
                       HTTPApplicationService, HTTPDomainService, HTTPInfraService, api, request, subscribe, publish,
                       message_pub, get, post, head, put, patch, delete, options, trace)
from .model import (Entity, Value, Aggregate, Factory, Repository)
from .bus import Bus
from .registry import Registry
from .utils import log
from .sql import PostgresStore, cursor, dict_cursor, nt_cursor, transaction

log.setup()



from asyncio import Future, get_event_loop, coroutine, iscoroutinefunction
from functools import wraps
import json

from again.utils import unique_hex
from aiohttp.web import Response

from .utils.ordered_class_member import OrderedClassMembers

# Service Client decorators


def make_request(func, self, args, kwargs, method):
    params = func(self, *args, **kwargs)
    entity = params.pop('entity', None)
    app_name = params.pop('app_name', None)
    self = params.pop('self')
    response = yield from self._send_http_request(app_name, method, entity, params)
    return response


def get_decorated_fun(method, path, required_params):
    def decorator(func):
        @wraps(func)
        def f(self, *args, **kwargs):
            if isinstance(self, HTTPServiceClient):
                return make_request(func, self, args, kwargs, method)
            elif isinstance(self, HTTPApplicationService):
                if required_params is not None:
                    req = args[0]
                    query_params = req.GET
                    params = required_params
                    if not isinstance(required_params, list):
                        params = [required_params]
                    if any(map(lambda x: x not in query_params, params)):
                        return Response(status=400, content_type='application/json',
                                        body=json.dumps({'error': 'Required params not found'}).encode())
                wrapped_func = func
                if not iscoroutinefunction(func):
                    wrapped_func = coroutine(func)
                return (yield from wrapped_func(self, *args, **kwargs))

        f.is_http_method = True
        f.method = method
        f.paths = path
        if not isinstance(path, list):
            f.paths = [path]
        return f

    return decorator


def get(path=None, required_params=None):
    return get_decorated_fun('get', path, required_params)


def head(path=None, required_params=None):
    return get_decorated_fun('head', path, required_params)


def options(path=None, required_params=None):
    return get_decorated_fun('options', path, required_params)


def patch(path=None, required_params=None):
    return get_decorated_fun('patch', path, required_params)


def post(path=None, required_params=None):
    return get_decorated_fun('post', path, required_params)


def put(path=None, required_params=None):
    return get_decorated_fun('put', path, required_params)


def trace(path=None, required_params=None):
    return get_decorated_fun('put', path, required_params)


def delete(path=None, required_params=None):
    return get_decorated_fun('delete', path, required_params)


def subscribe(func):
    """
    use to listen for publications from a specific endpoint of a service,
    this method receives a publication from a remote service
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    wrapper.is_subscribe = True
    return wrapper


def request(func):
    """
    use to request an api call from a specific endpoint
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        params = func(self, *args, **kwargs)
        self = params.pop('self', None)
        entity = params.pop('entity', None)
        app_name = params.pop('app_name', None)
        request_id = unique_hex()
        params['request_id'] = request_id
        future = self._send_request(app_name, endpoint=func.__name__, entity=entity, params=params)
        return future

    wrapper.is_request = True
    return wrapper


def message_sub(func):
    """
    use to listen for publications from a specific endpoint of a service,
    this method receives a publication from a remote service
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        params = func(self, *args, **kwargs)
        entity = params.pop('entity')
        try:
            app_name = params.pop('app_name')
        except KeyError:
            raise RuntimeError('App name must be specified')
        self._send_message_sub(app_name, endpoint=func.__name__, entity=entity)

    wrapper.is_directed_subscribe = True
    return wrapper


# Service Host Decorators

def publish(func):
    """
    publish the return value of this function as a message from this endpoint
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs):  # outgoing
        payload = func(self, *args, **kwargs)
        self._publish(func.__name__, payload)
        return None

    wrapper.is_publish = True

    return wrapper


def message_pub(func):
    """
    publish the return value of this function as a message from this endpoint
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs):  # outgoing
        payload = func(self, *args, **kwargs)
        entity = payload.pop('entity')
        self._message(func.__name__, payload, entity)
        return None

    wrapper.is_publish = True

    return wrapper


def api(func):  # incoming
    """
    provide a request/response api
    receives any requests here and return value is the response
    all functions must have the following signature
        - request_id
        - entity (partition/routing key)
        followed by kwargs
    """

    @coroutine
    @wraps(func)
    def wrapper(*args, **kwargs):
        self = args[0]
        rid = kwargs.pop('request_id')
        entity = kwargs.pop('entity')
        from_id = kwargs.pop('from_id')
        wrapped_func = func
        result = None
        error = None
        if not iscoroutinefunction(func):
            wrapped_func = coroutine(func)
        try:
            if len(kwargs):
                result = yield from wrapped_func(self, **kwargs)
            else:
                result = yield from wrapped_func()
        except BaseException as e:
            error = str(e)

        return self._make_response_packet(request_id=rid, from_id=from_id, entity=entity, result=result, error=error)

    wrapper.is_api = True
    return wrapper


class _Service:
    _PUB_PKT_STR = 'publish'
    _REQ_PKT_STR = 'request'
    _RES_PKT_STR = 'response'
    _MSG_PUB_PKT_STR = 'message_pub'
    _MSG_SUB_PKT_STR = 'message_sub'

    def __init__(self, service_name, service_version):
        self._service_name = service_name
        self._service_version = service_version
        self._bus = None

    @property
    def name(self):
        return self._service_name

    @property
    def version(self):
        return self._service_version

    @property
    def properties(self):
        return self.name, self.version

    @property
    def bus(self):
        return self._bus

    @bus.setter
    def bus(self, bus):
        self._bus = bus

    @staticmethod
    def time_future(future:Future, timeout:int):
        def timer_callback(f):
            if not f.done() and not f.cancelled():
                f.set_exception(TimeoutError())

        get_event_loop().call_later(timeout, timer_callback, future)


class TCPServiceClient(_Service):
    REQUEST_TIMEOUT_SECS = 10

    def __init__(self, service_name, service_version):
        super(TCPServiceClient, self).__init__(service_name, service_version)
        self._pending_requests = {}

    def _send_request(self, app_name, endpoint, entity, params):
        packet = self._make_packet(app_name, _Service._REQ_PKT_STR, endpoint, params, entity)
        future = Future()
        request_id = params['request_id']
        self._pending_requests[request_id] = future
        self._bus.send(packet)
        _Service.time_future(future, TCPServiceClient.REQUEST_TIMEOUT_SECS)
        return future

    def _send_message_sub(self, app_name, endpoint, entity):
        packet = self._make_packet(app_name, _Service._MSG_SUB_PKT_STR, endpoint, None, entity)
        self._bus.send(packet)

    def process_packet(self, packet):
        if packet['type'] == _Service._RES_PKT_STR:
            self._process_response(packet)
        elif packet['type'] == _Service._PUB_PKT_STR:
            self._process_publication(packet)
        else:
            print('Invalid packet', packet)

    def _process_response(self, packet):
        payload = packet['payload']
        request_id = payload['request_id']
        has_result = 'result' in payload
        has_error = 'error' in payload
        future = self._pending_requests.pop(request_id)
        if has_result:
            future.set_result(payload['result'])
        elif has_error:
            exception = RequestException()
            exception.error = payload['error']
            future.set_exception(exception)
        else:
            print('Invalid response to request:', packet)

    def _process_publication(self, packet):
        endpoint = packet['endpoint']
        func = getattr(self, endpoint)
        func(**packet['payload'])

    def _make_packet(self, app_name, packet_type, endpoint, params, entity):
        packet = {'pid': unique_hex(),
                  'app': app_name,
                  'service': self.name,
                  'version': self.version,
                  'entity': entity,
                  'endpoint': endpoint,
                  'type': packet_type,
                  'payload': params}
        return packet


class _ServiceHost(_Service):
    def __init__(self, service_name, service_version, host_ip, host_port):
        super(_ServiceHost, self).__init__(service_name, service_version)
        self._ip = host_ip
        self._port = host_port
        self._ronin = False

    def is_for_me(self, service, version):
        return service == self.name and int(version) == self.version

    @property
    def socket_address(self):
        return self._ip, self._port

    @property
    def ronin(self):
        return self._ronin

    @ronin.setter
    def ronin(self, value:bool):
        self._ronin = value


class _TCPServiceHost(_ServiceHost):
    def __init__(self, service_name, service_version, host_ip, host_port):
        # TODO: to be multi-tenant make app_name a list
        super(_TCPServiceHost, self).__init__(service_name, service_version, host_ip, host_port)

    def _publish(self, publication_name, payload):
        packet = self._make_publish_packet(_Service._PUB_PKT_STR, publication_name, payload)
        self._bus.send(packet)

    def _message(self, message_name, payload, entity):
        packet = self._make_publish_packet(_Service._MSG_PUB_PKT_STR, message_name, payload)
        packet['entity'] = entity
        self._bus.send(packet)

    def _make_response_packet(self, request_id: str, from_id: str, entity: str, result: object, error: object):
        if error:
            payload = {'request_id': request_id, 'error': error}
        else:
            payload = {'request_id': request_id, 'result': result}

        packet = {'pid': unique_hex(),
                  'to': from_id,
                  'entity': entity,
                  'type': _Service._RES_PKT_STR,
                  'payload': payload}
        return packet

    def _make_publish_packet(self, packet_type: str, publication_name: str, payload: dict):
        packet = {'service': self.name,
                  'version': self.version,
                  'endpoint': publication_name,
                  'type': packet_type,
                  'payload': payload}
        return packet


class RequestException(Exception):
    pass


class _HTTPServiceHost(_ServiceHost, metaclass=OrderedClassMembers):
    def __init__(self, service_name, service_version, host_ip, host_port, ssl_context=None):
        super(_HTTPServiceHost, self).__init__(service_name, service_version, host_ip, host_port)
        self._ssl_context = ssl_context

    @property
    def ssl_context(self):
        return self._ssl_context


class TCPApplicationService(_TCPServiceHost):
    pass


class TCPDomainService(_TCPServiceHost):
    pass


class TCPInfraService(_TCPServiceHost):
    pass


class HTTPApplicationService(_HTTPServiceHost):
    pass


class HTTPDomainService(_HTTPServiceHost):
    pass


class HTTPInfraService(_HTTPServiceHost):
    pass


class HTTPServiceClient(_Service):
    def __init__(self, service_name, service_version):
        super(HTTPServiceClient, self).__init__(service_name, service_version)

    def _send_http_request(self, app_name, method, entity, params):
        response = yield from self._bus.send_http_request(app_name, self.name, self.version, method, entity,
                                                          params)
        return response



from distutils.core import setup

setup(name='vyked',
      version='1.2.38',
      author='Kashif Razzaqui, Ankit Chandawala',
      author_email='kashif.razzaqui@gmail.com, ankitchandawala@gmail.com',
      url='https://github.com/kashifrazzaqui/vyked',
      description='A micro-service framework for Python',
      packages=['vyked', 'vyked.utils'], requires=['again', 'aiohttp', 'jsonstreamer', 'setproctitle', 'aiopg'])




import asyncio
from functools import partial
import os
import signal

from again.utils import unique_hex
import aiohttp
from aiohttp.web import Application, Response

from .jsonprotocol import ServiceHostProtocol, ServiceClientProtocol
from .registryclient import RegistryClient
from .services import TCPServiceClient, HTTPServiceClient, HTTPApplicationService

PUB_STORE = os.path.join(os.curdir, 'publish.store')


class Bus:
    def __init__(self, registry_host:str, registry_port:int):
        self._registry_host = registry_host
        self._registry_port = registry_port
        self._loop = asyncio.get_event_loop()
        self._client_protocols = {}
        self._service_clients = []
        self._death_listeners = set()
        self._pending_requests = []
        # TODO : replace with shelve
        self._unacked_publish = {}
        self._tcp_host = None
        self._http_host = None
        self._tcp_server = None
        self._http_server = None
        self._host_id = unique_hex()

    def require(self, args):
        for each in args:
            if isinstance(each, (TCPServiceClient, HTTPServiceClient)):
                each.bus = self
                self._service_clients.append(each)

    def add_death_listener(self, service:str, version:str):
        self._death_listeners.add((service, version))

    def serve_tcp(self, service_host):
        self._tcp_host = service_host

    def serve_http(self, service_host):
        self._http_host = service_host

    def send(self, packet:dict):
        packet['from'] = self._host_id
        func = getattr(self, '_' + packet['type'] + '_sender')
        func(packet)

    def send_http_request(self, app, service, version, method, entity, params):
        path = params.pop('path')
        query_params = params.pop('params', {})
        if app is not None:
            query_params['app'] = app
        query_params['version'] = version
        query_params['service'] = service
        data = params.pop('data', None)
        headers = params.pop('headers', None)
        cookies = params.pop('cookies', None)
        files = params.pop('files', None)
        auth = params.pop('auth', None)
        allow_redirects = params.pop('allow_redirects', True)
        max_redirects = params.pop('max_redirects', 10)
        encoding = params.pop('encoding', 'utf-8')
        http_version = params.pop('version', aiohttp.HttpVersion11)
        compress = params.pop('compress', None)
        chunked = params.pop('chunked', None)
        expect100 = params.pop('expect100', False)
        connector = params.pop('connector', None)
        loop = params.pop('loop', None)
        read_until_eof = params.pop('read_until_eof', True)
        request_class = params.pop('request_class', None)
        response_class = params.pop('response_class', None)
        host, port, node_id = self._registry_client.resolve(service, version, entity)
        # TODO : find a better method create the url
        url = 'http://{}:{}{}'.format(host, port, path)
        response = yield from aiohttp.request(method, url, params=query_params, data=data, headers=headers,
                                              cookies=cookies,
                                              files=files, auth=auth, allow_redirects=allow_redirects,
                                              max_redirects=max_redirects, encoding=encoding, version=http_version,
                                              compress=compress, chunked=chunked, expect100=expect100,
                                              connector=connector,
                                              loop=loop, read_until_eof=read_until_eof, request_class=request_class,
                                              response_class=response_class)
        return response

    def _request_sender(self, packet: dict):
        """
        sends a request to a server from a ServiceClient
        auto dispatch method called from self.send()
        """
        self._pending_requests.append(packet)
        self._clear_request_queue()

    def _message_sub_sender(self, packet: dict):
        packet['ip'], packet['port'] = self._tcp_host.socket_address
        self._registry_client.subscribe_for_message(packet)

    def _publish(self, future, packet):
        def send_publish_packet(publish_packet, f):
            transport, protocol = f.result()
            protocol.send(publish_packet)
            transport.close()

        def fun(fut):
            for node in fut.result():
                packet['to'] = node['node_id']
                pid = unique_hex()
                packet['pid'] = pid
                self._unacked_publish[pid] = packet
                coro = self._loop.create_connection(self._host_factory, node['ip'], node['port'])
                connect_future = asyncio.async(coro)
                connect_future.add_done_callback(partial(send_publish_packet, packet))

        future.add_done_callback(fun)

    def _publish_sender(self, packet: dict):
        """
        auto dispatch method called from self.send()
        """
        app, service, version, endpoint = packet['app'], packet['service'], packet['version'], packet['endpoint']
        future = self._registry_client.resolve_publication(service, version, endpoint)
        self._publish(future, packet)

    def _message_pub_sender(self, packet: dict):
        app, service, version, endpoint, entity = packet['app'], packet['service'], packet['version'], packet[
            'endpoint'], packet['entity']
        future = self._registry_client.resolve_message_publication(service, version, endpoint, entity)
        self._publish(future, packet)

    def host_receive(self, packet: dict, protocol: ServiceHostProtocol):
        if packet['type'] == 'ping':
            self._handle_ping(packet, protocol)
        elif packet['type'] == 'ack':
            pid = packet['pid']
            self._unacked_publish.pop(pid)
        elif packet['type'] == 'publish':
            client = [sc for sc in self._service_clients if (
                sc.name == packet['service'] and sc.version == packet['version'])][0]
            func = getattr(client, packet['endpoint'])
            func(packet['payload'])
            self.send_ack(protocol, packet['pid'])
        else:
            if self._tcp_host.is_for_me(packet['service'], packet['version']):
                func = getattr(self, '_' + packet['type'] + '_receiver')
                func(packet, protocol)
            else:
                print('wrongly routed packet: ', packet)

    def _request_receiver(self, packet, protocol):
        api_fn = getattr(self._tcp_host, packet['endpoint'])
        if api_fn.is_api:
            from_node_id = packet['from']
            entity = packet['entity']
            future = asyncio.async(api_fn(from_id=from_node_id, entity=entity, **packet['payload']))

            def send_result(f):
                result_packet = f.result()
                protocol.send(result_packet)

            future.add_done_callback(send_result)
        else:
            print('no api found for packet: ', packet)

    def client_receive(self, service_client:TCPServiceClient, packet:dict):
        service_client.process_packet(packet)

    def _stop(self, signame:str):
        print('\ngot signal {} - exiting'.format(signame))
        self._loop.stop()

    def _host_factory(self):
        return ServiceHostProtocol(self)

    def _client_factory(self):
        return ServiceClientProtocol(self)

    def is_tcp_ronin(self):
        return self._tcp_host and not self._tcp_host.ronin

    def is_http_ronin(self):
        return self._http_host and not self._http_host.ronin

    def start(self):
        self._set_process_name()
        self._loop.add_signal_handler(getattr(signal, 'SIGINT'), partial(self._stop, 'SIGINT'))
        self._loop.add_signal_handler(getattr(signal, 'SIGTERM'), partial(self._stop, 'SIGTERM'))

        self._tcp_server = self._create_tcp_service_host()
        self._http_server = self._create_http_service_host()
        if self.is_tcp_ronin() or self.is_http_ronin():
            self._setup_registry_client()

        # TODO: All the ronin conditional logic needs refactor and completion
        if self.is_tcp_ronin():
            tcp_host_ip, tcp_host_port = self._tcp_host.socket_address
            self._registry_client.register_tcp(self._service_clients, tcp_host_ip, tcp_host_port,
                                               *self._tcp_host.properties)

        if self.is_http_ronin():
            ip, port = self._http_host.socket_address
            self._registry_client.register_http(self._service_clients, ip, port,
                                                *self._http_host.properties)

        if self._tcp_server:
            print('Serving TCP on {}'.format(self._tcp_server.sockets[0].getsockname()))
        if self._http_server:
            print('Serving HTTP on {}'.format(self._http_server.sockets[0].getsockname()))
        print("Event loop running forever, press CTRL+c to interrupt.")
        print("pid %s: send SIGINT or SIGTERM to exit." % os.getpid())

        try:
            self._loop.run_forever()
        except Exception as e:
            print(e)
        finally:
            if self._tcp_server:
                self._tcp_server.close()
                self._loop.run_until_complete(self._tcp_server.wait_closed())

            if self._http_server:
                self._http_server.close()
                self._loop.run_until_complete(self._http_server.wait_closed())

            self._loop.close()

    def registration_complete(self):
        for service, version in self._death_listeners:
            self._registry_client.add_service_death_listener(service, version)
        f = self._create_service_clients()

        def fun(f):
            if self._tcp_host:
                self._clear_request_queue()

        f.add_done_callback(fun)

    def _create_tcp_service_host(self):
        if self._tcp_host:
            host_ip, host_port = self._tcp_host.socket_address
            host_coro = self._loop.create_server(self._host_factory, host_ip, host_port)
            return self._loop.run_until_complete(host_coro)

    def verify(self, func):
        def verified_func(*args, **kwargs):
            query_dict = args[0].GET
            if isinstance(self._http_host, HTTPApplicationService):
                return func(*args, **kwargs)
            if 'service' in query_dict and 'version' in query_dict:
                if self._http_host.is_for_me(query_dict['service'], query_dict['version']):
                    return func(*args, **kwargs)
                else:
                    return Response(status=421, body="421 wrongly routed request".encode())
            else:
                return Response(status=400, body="400 bad request".encode())

        return verified_func

    def _create_http_service_host(self):
        if self._http_host:
            host_ip, host_port = self._http_host.socket_address
            ssl_context = self._http_host.ssl_context
            app = Application(loop=self._loop)
            for each in self._http_host.__ordered__:
                fn = getattr(self._http_host, each)
                if callable(fn) and getattr(fn, 'is_http_method', False):
                    for path in fn.paths:
                        app.router.add_route(fn.method, path, self.verify(fn))
            handler = app.make_handler()
            http_coro = self._loop.create_server(handler, host_ip, host_port, ssl=ssl_context)
            return self._loop.run_until_complete(http_coro)

    def _create_service_clients(self):
        futures = []
        for sc in self._service_clients:
            for host, port, node_id in self._registry_client.get_all_addresses(sc.properties):
                coro = self._loop.create_connection(self._client_factory, host, port)
                future = asyncio.async(coro)
                future.add_done_callback(partial(self._service_client_connection_callback, sc, node_id))
                futures.append(future)
        return asyncio.gather(*futures, return_exceptions=False)

    def _service_client_connection_callback(self, sc, node_id, future):
        transport, protocol = future.result()
        protocol.set_service_client(sc)
        self._client_protocols[node_id] = protocol

    def _setup_registry_client(self):
        self._registry_client = RegistryClient(self._loop, self._registry_host, self._registry_port, self)
        self._registry_client.connect()

    @staticmethod
    def _create_json_service_name(app, service, version):
        return {'app': app, 'service': service, 'version': version}

    def _handle_ping(self, packet, protocol):
        pong_packet = self._make_pong_packet(packet['node_id'], packet['count'])
        protocol.send(pong_packet)

    def _make_pong_packet(self, node_id, count):
        packet = {'type': 'pong', 'node_id': node_id, 'count': count}
        return packet

    def _clear_request_queue(self):
        for packet in self._pending_requests:
            app, service, version, entity = packet['app'], packet['service'], packet['version'], packet['entity']
            node = self._registry_client.resolve(service, version, entity)
            if node is not None:
                node_id = node[2]
                client_protocol = self._client_protocols[node_id]
                packet['to'] = node_id
                client_protocol.send(packet)
                self._pending_requests.remove(packet)

    @staticmethod
    def send_ack(protocol, pid):
        packet = {'type': 'ack', 'pid': pid}
        protocol.send(packet)

    def _set_process_name(self):
        from setproctitle import setproctitle

        if self._tcp_host:
            setproctitle('{}_{}'.format(self._tcp_host.name, self._tcp_host.version))
        elif self._http_host:
            setproctitle('{}_{}'.format(self._http_host.name, self._http_host.version))


if __name__ == '__main__':
    REGISTRY_HOST = '127.0.0.1'
    REGISTRY_PORT = 4500
    HOST_IP = '127.0.0.1'
    HOST_PORT = 8000
    bus = Bus(REGISTRY_HOST, REGISTRY_PORT)
    bus.start()
